class TensorShapeMismatchError(Exception):
    """Mismatch in the shape of the inputs."""


class LinalgOperationInputError(Exception):
    """Invalid input for this operation."""


class LinalgMissingInputError(Exception):
    """Invalid input for this operation."""


class LinalgInvalidTensorOperationError(Exception):
    """
    Tensors don't support this operation. Try to call this using
    the arrays of the tensorial inputs.
    """



from abc import abstractmethod
from typing import Iterable, Callable, Tuple, List
import numpy as np
from numpy import ndarray


__all__ = ["GeneticAlgorithm", "BinaryGeneticAlgorithm"]


def even(n):
    return n % 2 == 0


def odd(n):
    return not even(n)


class GeneticAlgorithm:
    """
    Base class for Genetic Algorithms (GA). Use this as a base
    class to your custom implementation of a GA.

    The class has 4 abstract methods wich upon being implemented, yield
    a working genetic algorithm. These are :func:`populate`, :func:`decode`,
    :func:`crossover`, :func:`mutate` and :func:`select`.
    See :class:`~neumann.optimize.ga.BinaryGeneticAlgorithm` for an example.

    Parameters
    ----------
    fnc : Callable
        The function to evaluate. It is assumed, that the function expects and N
        number of scalar arguments as a 1d iterable.
    ranges : Iterable
        Ranges for each scalar argument to the objective function.
    length : int, Optional
        Chromosome length. The higher the value, the more precision. Default is 5.
    p_c : float, Optional
        Probability of crossover. Default is 1.
    p_m : float, Optional
        Probability of mutation. Default is 0.2.
    nPop : int, Optional
        The size of the population. Default is 100.
    maxiter : int, Optional
        The maximum number of iterations. Default is 200.
    miniter : int, Optional
        The minimum number of iterations. Default is 100.
    elitism : float or int, Optional
        Default is 1

    Note
    ----
    Be cautious what you use a genetic algorithm for. Like all metahauristic methods, a
    genetic algorithm can be wery demanding on the computational side. If the objective
    unction takes a lot of time to evaluate, it is probably not a good idea to use a heuristic
    approach, unless you have a dedicated solver that is able to run efficiently for a large
    number of problems.

    See also
    --------
    :class:`~neumann.optimize.ga.BinaryGeneticAlgorithm`
    """

    def __init__(
        self,
        fnc: Callable,
        ranges: Iterable,
        *,
        length: int = 5,
        p_c: float = 1,
        p_m: float = 0.2,
        nPop: int = 100,
        **kwargs
    ):
        super().__init__()
        self.fnc = fnc
        self.ranges = np.array(ranges)
        self.dim = getattr(fnc, "dimension", self.ranges.shape[0])
        self.length = length
        self.p_c = p_c
        self.p_m = p_m

        # Second half of the population is used as a pool to make parents.
        # This assumes that population size is a multiple of 4.
        if odd(nPop):
            nPop += 1
        if odd(int(nPop / 2)):
            nPop += 2
        assert nPop % 4 == 0
        assert nPop >= 4

        self.nPop = nPop
        self._genotypes = None
        self._fittness = None
        self.reset()
        self._set_solution_params(**kwargs)

    @property
    def genotypes(self):
        """
        Returnes the genotypes of the population.
        """
        return self._genotypes

    @genotypes.setter
    def genotypes(self, value):
        """
        Sets the genotypes of the population.
        """
        self._genotypes = value
        self.phenotypes = self.decode(self._genotypes)

    def reset(self):
        """
        Resets the solver.
        """
        self._evolver = self.evolver()
        self._evolver.send(None)

    def _set_solution_params(
        self, tol=1e-12, maxiter=200, miniter=100, elitism=1, **kwargs
    ):
        self.tol = tol
        self.maxiter = np.max([miniter, maxiter])
        self.miniter = np.min([miniter, maxiter])
        self.elitism = elitism

    def evolver(self):
        """
        Returns a generator that can be used to manually control evolutions.
        """
        self.genotypes = self.populate()
        _ = yield
        while True:
            self.genotypes = self.populate(
                self.select(self._genotypes, self.phenotypes)
            )
            yield self._genotypes

    def evolve(self, cycles=1):
        """
        Performs one cycle of evolution.
        """
        for _ in range(cycles):
            next(self._evolver)
        return self.genotypes

    def criteria(self) -> bool:
        value = yield
        while True:
            _value = yield
            yield abs(value - _value) < self.tol
            value = _value

    def solve(self, reset: bool = False, returnlast: bool = False, **kwargs):
        """
        Solves the problem and returns the best phenotype.
        """
        if reset:
            self.reset()
        self._set_solution_params(**kwargs)
        criteria = self.criteria()
        criteria.send(None)
        criteria.send(self.fnc(self.best_phenotype()))
        finished = False
        nIter = 0
        while (not finished and nIter < self.maxiter) or (nIter < self.miniter):
            next(self._evolver)
            finished = criteria.send(self.fnc(self.best_phenotype()))
            next(criteria)
            nIter += 1
        self.nIter = nIter
        return self.best_phenotype(lastknown=returnlast)

    def fittness(self, phenotypes: ndarray = None) -> ndarray:
        """
        Returns the actual fittness values of the population, or the fittness
        of the population described by the argument `phenotypes`.
        """
        if phenotypes is not None:
            self._fittness = np.array([self.fnc(x) for x in phenotypes], dtype=float)
        return self._fittness

    def best_phenotype(self, lastknown: bool = False) -> ndarray:
        """
        Returns the best phenotype.

        Parameters
        ----------
        lastknown : bool, Optional
            If True, the last evaluation is used. If False, the phenotypes
            are evaluated before selecting the best. In this case the results
            are not stored. Default is False.
        """
        if lastknown:
            fittness = self._fittness
        else:
            fittness = self.fittness(self.phenotypes)
        best = np.argmin(fittness)
        return self.phenotypes[best]

    def divide(self, fittness: ndarray = None) -> Tuple[List]:
        """
        Divides population to elit and others, and returns the corresponding
        index arrays.

        Parameters
        ----------
        fittness : numpy.ndarray, Optional
            Fittness values. If not provided, values from the latest
            evaluation are used. Default is None.

        Returns
        -------
        list
            Indices of the members of the elite.
        list
            Indices of the members of the others.
        """
        fittness = self.fittness() if fittness is None else fittness
        assert fittness is not None, "No available fittness data detected."
        if self.elitism < 1:
            argsort = np.argsort(fittness)
            elit = argsort[: int(self.nPop * self.elitism)]
            others = argsort[int(self.nPop * self.elitism) :]
        elif self.elitism > 1:
            argsort = np.argsort(fittness)
            elit = argsort[: self.elitism]
            others = argsort[self.elitism :]
        else:
            elit = []
            others = list(range(self.nPop))
        return list(elit), others

    @classmethod
    def random_parents_generator(cls, genotypes: ndarray = None):
        """
        Yields random pairs from a list of genotypes.

        The implemantation assumes that the length of the input array
        is a multiple of 2.

        Parameters
        ----------
        genotypes : numpy.ndarray
            Genotypes of the parents as a 2d integer array.

        Yields
        ------
        numpy.ndarray
            The first parent.
        numpy.ndarray
            The second parent.
        """
        n = len(genotypes)
        assert n % 2 == 0, "'n' must be a multiple of 2"
        pool = np.full(n, True)
        nPool = n
        while nPool > 2:
            where = np.argwhere(pool == True).flatten()
            nPool = len(where)
            pair = np.random.choice(where, 2, replace=False)
            parent1 = genotypes[pair[0]]
            parent2 = genotypes[pair[1]]
            pool[pair] = False
            yield parent1, parent2

    @abstractmethod
    def populate(self, genotypes: ndarray = None):
        """
        Ought to produce a pool of phenotypes.
        """
        ...

    @abstractmethod
    def decode(self, genotypes: ndarray = None):
        """
        Turns genotypes into phenotypes.
        """
        ...

    @abstractmethod
    def crossover(
        self, parent1: ndarray = None, parent2: ndarray = None
    ) -> Tuple[ndarray]:
        """
        Takes in two parents, returns two offspring. You'd probably want to use it inside
        the populator.
        """
        ...

    @abstractmethod
    def mutate(self, child: ndarray = None) -> ndarray:
        """
        Takes a child in, returns a mutant.
        """
        ...

    @abstractmethod
    def select(self, genotypes: ndarray = None, phenotypes: ndarray = None):
        """
        Ought to implement dome kind of selection mechanism, eg. a roulette wheel,
        tournament or other.
        """
        ...


class BinaryGeneticAlgorithm(GeneticAlgorithm):
    """
    An implementation of a Binary Genetic Algorithm (BGA) for finding
    minimums of real valued unconstrained problems of continuous variables
    in n-dimensional vector spaces.

    In other words, it solves the following problem:

    .. math::
        :nowrap:

        \\begin{eqnarray}
            & minimize&  \quad  f(\mathbf{x}) \quad in \quad \mathbf{x} \in \mathbf{R}^n.
        \\end{eqnarray}

    Parameters
    ----------
    fnc : Callable
        The fittness function.
    ranges : Iterable
        sequence of pairs of limits for each variable
    length : int, Optional
        Chromosome length (string length). Default is 5.
    p_c : float, Optional
        Probability of crossover, 0 <= p_c <= 1. Default is 1.
    p_m : float, Optional
        Probability of mutation, 0 <= p_m <= 1. Default is 0.2.
    nPop : int, Optional
        Number of members in the population. Default is 100.
    elitism : float or integer, Optional
        Value to control elitism. Default is 1.

    Examples
    --------
    Find the minimizer of the Rosenbrock function.
    The exact value of the solution is x = [1.0, 1.0].

    >>> from neumann.optimize import BinaryGeneticAlgorithm
    >>> def Rosenbrock(x):
    ...     a, b = 1, 100
    ...     return (a-x[0])**2 + b*(x[1]-x[0]**2)**2
    >>> ranges = [[-10, 10], [-10, 10]]
    >>> BGA = BinaryGeneticAlgorithm(Rosenbrock, ranges, length=12, nPop=200)
    >>> _ = BGA.solve()
    >>> x = BGA.best_phenotype()
    >>> fx = Rosenbrock(x)
    ...

    The following code prints the history using the `evolve` generator of
    the object

    >>> from neumann.optimize import BinaryGeneticAlgorithm
    >>> import matplotlib.pyplot as plt
    >>> def Rosenbrock(x):
    ...     a, b = 1, 100
    ...     return (a-x[0])**2 + b*(x[1]-x[0]**2)**2
    >>> ranges = [[-10, 10], [-10, 10]]
    >>> BGA = BinaryGeneticAlgorithm(Rosenbrock, ranges, length=12, nPop=200)
    >>> _ = [BGA.evolve(1) for _ in range(100)]
    >>> x = BGA.best_phenotype()
    >>> fx = Rosenbrock(x)
    ...
    """

    def populate(self, genotypes: ndarray = None):
        """
        Populates the model from a list of genotypes as seeds.
        """
        nPop = self.nPop
        if genotypes is None:
            poolshape = (int(nPop / 2), self.dim * self.length)
            genotypes = np.random.randint(2, size=poolshape)
        else:
            poolshape = genotypes.shape
        nParent = poolshape[0]
        if nParent < nPop:
            offspring = []
            g = self.random_parents_generator(genotypes)
            try:
                while (len(offspring) + nParent) < nPop:
                    parent1, parent2 = next(g)
                    offspring.extend(self.crossover(parent1, parent2))
                genotypes = np.vstack([genotypes, offspring])
            except Exception:
                raise RuntimeError
        return genotypes

    def decode(self, genotypes: ndarray = None) -> ndarray:
        """
        Decodes the genotypes to phenotypes.
        """
        span = 2**self.length - 2**0
        genotypes = genotypes.reshape((self.nPop, self.dim, self.length))
        precisions = [
            (self.ranges[d, -1] - self.ranges[d, 0]) / span for d in range(self.dim)
        ]
        phenotypes = np.sum(
            [genotypes[:, :, i] * 2**i for i in range(self.length)], axis=0
        ).astype(float)
        for d in range(self.dim):
            phenotypes[:, d] *= precisions[d]
            phenotypes[:, d] += self.ranges[d, 0]
        return phenotypes

    def crossover(
        self, parent1: ndarray = None, parent2: ndarray = None, nCut: int = None
    ) -> Tuple[ndarray]:
        """
        Performs crossover on the parents `parent1` and `parent2`,
        using an `nCut` number of cuts and returns two childs.
        """
        if np.random.rand() > self.p_c:
            return parent1, parent2

        if nCut is None:
            nCut = np.random.randint(1, self.dim * self.length - 1)

        cuts = [0, self.dim * self.length]
        p = np.random.choice(range(1, self.length * self.dim - 1), nCut, replace=False)
        cuts.extend(p)
        cuts = np.sort(cuts)

        child1 = np.zeros(self.dim * self.length, dtype=int)
        child2 = np.zeros(self.dim * self.length, dtype=int)

        randBool = np.random.rand() > 0.5
        for i in range(nCut + 1):
            if (i % 2 == 0) == randBool:
                child1[cuts[i] : cuts[i + 1]] = parent1[cuts[i] : cuts[i + 1]]
                child2[cuts[i] : cuts[i + 1]] = parent2[cuts[i] : cuts[i + 1]]
            else:
                child1[cuts[i] : cuts[i + 1]] = parent2[cuts[i] : cuts[i + 1]]
                child2[cuts[i] : cuts[i + 1]] = parent1[cuts[i] : cuts[i + 1]]

        return self.mutate(child1), self.mutate(child2)

    def mutate(self, child: ndarray = None) -> ndarray:
        """
        Returns a mutated genotype. Children come in, mutants go out.
        """
        p = np.random.rand(self.dim * self.length)
        return np.where(p > self.p_m, child, 1 - child)

    def select(self, genotypes: ndarray = None, phenotypes: ndarray = None) -> ndarray:
        """
        Organizes a tournament and returns the genotypes of the winners.
        """
        fittness = self.fittness(phenotypes)
        winners, others = self.divide(fittness)
        while len(winners) < int(self.nPop / 2):
            candidates = np.random.choice(others, 3, replace=False)
            winner = np.argsort([fittness[ID] for ID in candidates])[0]
            winners.append(candidates[winner])
        return np.array([genotypes[w] for w in winners], dtype=float)



from typing import Iterable
from copy import deepcopy as dcopy

import numpy as np
from numpy import ndarray
import sympy as sy

from dewloosh.core.tools.alphabet import latinrange
from .frame import ReferenceFrame as Frame
from .abstract import AbstractTensor
from .top import tr_3333, tr_3333_jit
from .utils import is_hermitian, _transform_tensors2_multi


__all__ = ["Tensor", "Tensor2", "Tensor4", "Tensor2x3", "Tensor4x3"]


class Tensor(AbstractTensor):
    """
    A class to handle tensors.

    Parameters
    ----------
    args : tuple, Optional
        Positional arguments forwarded to `numpy.ndarray`.
    frame : numpy.ndarray, Optional
        The reference frame the vector is represented by its coordinates.
    kwargs : dict, Optional
        Keyword arguments forwarded to `numpy.ndarray`.

    Examples
    --------
    Import the necessary classes:

    >>> from neumann.linalg import Tensor, ReferenceFrame
    >>> from numpy.random import rand

    Create a Tensor of order 6 in a frame with random components

    >>> frame = ReferenceFrame(dim=3)
    >>> array = rand(3, 3, 3, 3, 3, 3)
    >>> A = Tensor(array, frame=frame)

    Get the tensor in the dual frame:

    >>> A_dual = A.dual()

    Create an other tensor, in this case a 5th-order one, and calculate their
    generalized dot product, which is a 9th-order tensor:

    >>> from neumann.linalg import dot
    >>> array = rand(3, 3, 3, 3, 3)
    >>> B = Tensor(array, frame=frame)
    >>> C = dot(A, B, axes=[0, 0])
    >>> assert C.rank == (A.rank + B.rank - 2)

    See Also
    --------
    :class:`~neumann.linalg.vector.Vector`
    :class:`~neumann.linalg.frame.ReferenceFrame`
    """

    _frame_cls_ = Frame
    _einsum_params_ = {}

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, **kwargs) -> bool:
        return is_hermitian(arr)

    @classmethod
    def _from_any_input(cls, *args, **kwargs) -> AbstractTensor:
        if cls._verify_input(*args, **kwargs):
            return cls(*args, **kwargs)
        else:
            if not Tensor._verify_input(*args, **kwargs):
                raise ValueError("Invalid input to Tensor class.")
            else:
                return Tensor(*args, **kwargs)

    def dual(self) -> "Tensor2":
        """
        Returns the tensor described in the dual (or reciprocal) frame.
        """
        a = self.transform_components(self.frame.Gram())
        return self.__class__(a, frame=self.frame.dual())

    def transform_components(self, Q: ndarray) -> ndarray:
        """
        Returns the components of the tensor transformed by the matrix Q.
        """
        r = self.rank
        arr = self.array
        args = [Q for _ in range(r)]
        if r not in self.__class__._einsum_params_:
            target = latinrange(r, start=ord("a"))
            source = latinrange(r, start=ord("a") + r)
            terms = [t + s for s, t in zip(source, target)]
            command = ",".join(terms) + "," + "".join(source)
            einsum_path = np.einsum_path(command, *args, arr, optimize="greedy")[0]
            self.__class__._einsum_params_[r] = (command, einsum_path)
        else:
            command, einsum_path = self.__class__._einsum_params_[r]
        return np.einsum(command, *args, arr, optimize=einsum_path)

    def show(self, target: Frame = None, *, dcm: ndarray = None) -> ndarray:
        """
        Returns the components in a target frame. If the target is
        `None`, the components are returned in the ambient frame.

        The transformation can also be specified with a proper DCM matrix.

        Parameters
        ----------
        target : numpy.ndarray, Optional
            Target frame.
        dcm : numpy.ndarray, Optional
            The DCM matrix of the transformation.

        Returns
        -------
        numpy.ndarray
            The components of the tensor in a specified frame, or
            the ambient frame, depending on the arguments.
        """
        if not isinstance(dcm, ndarray):
            if target is None:
                target = Frame(dim=self._array.shape[-1])
            dcm = self.frame.dcm(target=target)
        return self.transform_components(dcm)

    def orient(self, *args, **kwargs) -> "Tensor":
        """
        Orients the vector inplace. All arguments are forwarded to
        `orient_new`.

        Returns
        -------
        Vector
            The same vector the function is called upon.

        See Also
        --------
        :func:`orient_new`
        """
        fcls = self.__class__._frame_cls_
        dcm = fcls.eye(dim=len(self)).orient_new(*args, **kwargs).dcm()
        self.array = self.transform_components(dcm.T)
        return self

    def orient_new(self, *args, **kwargs) -> "Tensor":
        """
        Returns a transformed version of the instance.

        Returns
        -------
        Vector
            A new vector.

        See Also
        --------
        :func:`orient`
        """
        fcls = self.__class__._frame_cls_
        dcm = fcls.eye(dim=len(self)).orient_new(*args, **kwargs).dcm()
        array = self.transform_components(dcm.T)
        return self.__class__(array, frame=self.frame)

    def copy(self, deep: bool = False, name: str = None) -> "Tensor":
        """
        Returns a shallow or deep copy of this object, depending of the
        argument `deepcopy` (default is False).
        """
        if deep:
            return self.__class__(dcopy(self.array), name=name)
        else:
            return self.__class__(self.array, name=name)

    def deepcopy(self, name: str = None) -> "Tensor":
        """
        Returns a deep copy of the frame.
        """
        return self.copy(deep=True, name=name)


class Tensor2(Tensor):
    """
    A class to handle 2nd-order tensors. Some operations have dedicated implementations
    that provide higher performence utilizing implicit parallelization. Examples include
    the metric tensor, or the stress and strain tensors of elasticity.
    """

    _rank_ = 2

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, bulk: bool = False, **kwargs) -> bool:
        if bulk:
            return len(arr.shape) == 3 and is_hermitian(arr[0])
        else:
            return len(arr.shape) == 2 and is_hermitian(arr)

    def transform_components(self, Q: ndarray) -> ndarray:
        if len(Q.shape) == 3 and len(self.array.shape) == 3:
            return _transform_tensors2_multi(self.array, Q)
        else:
            return Q @ self.array @ Q.T


class Tensor2x3(Tensor2):
    ...


class Tensor4(Tensor):
    """
    A class to handle 4th-order tensors. Some operations have dedicated implementations
    that provide higher performence utilizing implicit parallelization. Examples include
    the piezo-optical tensor, the elasto-optical tensor, the flexoelectric tensor or the
    elasticity tensor.
    """

    _rank_ = 4

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, bulk: bool = False, **kwargs) -> bool:
        if bulk:
            return len(arr.shape) == 5 and is_hermitian(arr[0])
        else:
            return len(arr.shape) == 4 and is_hermitian(arr)


class Tensor4x3(Tensor):
    """
    A class for fourth-order tensors, where each index ranges from 1 to 3.

    Parameters
    ----------
    imap : dict, Optional
        An invertible index map for second-order tensors that assigns to each pair
        of indices a single index. The index map used to switch between 4d and 2d
        representation is inferred from this input. The default is the Voigt indicial map:
            0 : (0, 0)
            1 : (1, 1)
            2 : (2, 2)
            3 : (1, 2)
            4 : (0, 2)
            5 : (0, 1)
    symbolic : bool, Optional
        If True, the tensor is stored in symbolic form, and the components are stored as
        a `SymPy` matrix. Default is False.
    """

    __imap__ = {0: (0, 0), 1: (1, 1), 2: (2, 2), 3: (1, 2), 4: (0, 2), 5: (0, 1)}

    def __init__(self, *args, symbolic: bool = False, imap: dict = None, **kwargs):
        super().__init__(*args, **kwargs)

        if isinstance(imap, dict):
            self.__imap__ = imap

        if symbolic:
            self.transform_components = self._transform_sym_
            self.dtype = object

    @property
    def collapsed(self):
        if self._array is not None:
            return self._array.shape[-1] == 6
        else:
            raise ValueError("There is no data.")

    def expand(self) -> "Tensor4x3":
        """
        Changes the representation of the tensor to 4d.
        """
        if not self.collapsed:
            return self
        T = np.zeros((3, 3, 3, 3), dtype=self._array.dtype)
        m = self._array
        imap = self.imap()
        for ij, ijkl in imap.items():
            T[ijkl] = m[ij]
        self._array = T
        return self

    def collapse(self) -> "Tensor4x3":
        """
        Changes the representation of the tensor to 2d.
        """
        if self.collapsed:
            return self
        m = np.zeros((6, 6), dtype=self._array.dtype)
        T = self._array
        imap = self.imap()
        for ij, ijkl in imap.items():
            m[ij] = T[ijkl]
        self._array = m
        return self

    @classmethod
    def imap(cls, imap1d: dict = None) -> dict:
        """
        Returns a 2d-to-4d index map used to collapse or expand a tensor,
        based on the 1d-to-2d mapping of the class the function is called on,
        or on the first argument, if it is a suitable candidate for an
        index map.
        """
        if imap1d is None:
            imap1d = cls.__imap__
        indices = np.indices((6, 6))
        it = np.nditer([*indices], ["multi_index"])
        imap2d = dict()
        for _ in it:
            i, j = it.multi_index
            imap2d[(i, j)] = imap1d[i] + imap1d[j]
        return imap2d

    @classmethod
    def symbolic(
        cls, *args, base: str = "C_", as_matrix: bool = False, imap: dict = None
    ) -> Iterable:
        """
        Returns a symbolic representation of a 4th order 3x3x3x3 tensor.
        If the argument 'as_matrix' is True, the function returns a 6x6 matrix,
        that unfolds according to the argument 'imap', or if it's not provided,
        the index map of the class the function is called on. If 'imap' is
        provided, it must be a dictionary including exactly 6 keys and
        values. The keys must be integers in the integer range (0, 6), the
        values must be tuples on the integer range (0, 3).
        The default mapping is the Voigt indicial map:
            0 : (0, 0)
            1 : (1, 1)
            2 : (2, 2)
            3 : (1, 2)
            4 : (0, 2)
            5 : (0, 1)
        """
        res = np.zeros((3, 3, 3, 3), dtype=object)
        indices = np.indices((3, 3, 3, 3))
        it = np.nditer([*indices], ["multi_index"])
        for _ in it:
            p, q, r, s = it.multi_index
            if q >= p and s >= r:
                sinds = np.array([p, q, r, s], dtype=np.int16) + 1
                sym = sy.symbols(base + "_".join(sinds.astype(str)))
                res[p, q, r, s] = sym
                res[q, p, r, s] = sym
                res[p, q, s, r] = sym
                res[q, p, s, r] = sym
                res[r, s, p, q] = sym
                res[r, s, q, p] = sym
                res[s, r, p, q] = sym
                res[s, r, q, p] = sym
        if as_matrix:
            mat = np.zeros((6, 6), dtype=object)
            imap = cls.imap(imap) if imap is None else imap
            for ij, ijkl in imap.items():
                mat[ij] = res[ijkl]
            if "sympy" in args:
                res = sy.Matrix(mat)
            else:
                res = mat
        return res

    def transform_components(self, dcm: np.ndarray) -> ndarray:
        """
        Returns the components of the transformed numerical tensor, based on
        the provided direction cosine matrix.
        """
        if self.collapsed:
            self.expand()
            array = tr_3333_jit(self._array, dcm)
            self.collapse()
        else:
            array = tr_3333_jit(self._array, dcm)
        return array

    def _transform_sym_(self, dcm: np.ndarray) -> ndarray:
        """
        Returns the components of the transformed symbolic tensor, based on
        the provided direction cosine matrix.
        """
        if self.collapsed:
            self.expand()
            array = tr_3333(self._array, dcm, dtype=object)
            self.collapse()
        else:
            array = tr_3333(self._array, dcm, dtype=object)
        return array



from typing import Union
import numbers
import itertools

import numpy as np
from numpy import ndarray
from numba import njit, prange, guvectorize as guv
import sympy as sy
from sympy import symbols, Matrix

from dewloosh.core.tools.alphabet import latinrange

from .meta import TensorLike, ArrayWrapper, FrameLike
from .exceptions import LinalgOperationInputError, LinalgMissingInputError

__cache = True


__all__ = [
    "permutation_tensor",
    "dot",
    "cross",
    "is_rectangular_frame",
    "is_normal_frame",
    "is_orthonormal_frame",
    "is_independent_frame",
    "is_hermitian",
    "normalize_frame",
    "Gram",
    "dual_frame",
    "is_pos_def",
    "is_pos_semidef",
    "random_pos_semidef_matrix",
    "random_posdef_matrix",
    "inv_sym_3x3",
    "vpath",
    "det3x3",
    "det2x2",
    "inv2x2",
    "inv2x2u",
    "adj3x3",
    "inv3x3u",
    "inv3x3",
    "inv3x3_bulk",
    "inv3x3_bulk2",
    "normalize",
    "normalize2d",
    "norm",
    "norm2d",
    "linspace",
    "linspace1d",
    "inv",
]


def permutation_tensor(dim: int = 3) -> ndarray:
    """
    Returns the Levi-Civita pseudotensor for N dimensions.

    Parameters
    ----------
    N : int, Optional
        The number of dimensions. Default is 3.
    """
    arr = np.zeros(tuple([dim for _ in range(dim)]))
    mat = np.zeros((dim, dim), dtype=np.int32)
    for x in itertools.permutations(tuple(range(dim))):
        mat[:, :] = 0
        for i, j in zip(range(dim), x):
            mat[i, j] = 1
        arr[x] = int(np.linalg.det(mat))
    return arr


def dot(
    a: Union[TensorLike, ArrayWrapper],
    b: Union[TensorLike, ArrayWrapper],
    out: Union[TensorLike, ArrayWrapper] = None,
    frame: FrameLike = None,
    axes: Union[list, tuple] = None,
) -> Union[TensorLike, ndarray, numbers.Number]:
    """
    Returns the dot product (without complex conjugation) of two quantities. The behaviour
    coincides with NumPy when all inputs are arrays and generalizes when they are not,
    but all inputs must be either all arrays or all tensors of some kind. The operation for
    tensors of order 1 and 2 have dedicated implementations, for higher order tensors
    it generalizes to tensor contraction along specified axes.

    Parameters
    ----------
    a : TensorLike or ArrayLike
        A tensor or an array.
    b : TensorLike or ArrayLike
        A tensor or an array.
    out : ArrayLike, Optional
        Output argument. This must have the exact kind that would be returned if it was
        not used. See `numpy.dot` for the details. Only if all inputs are ArrayLike.
        Default is None.
    frame : FrameLike, Optinal
        The target frame of the output. Only if all inputs are TensorLike. If not specified,
        the returned tensor migh be returned in an arbitrary frame, depending on the inputs.
        Default is None.
    axes : tuple or list, Optional
        The indices along which contraction happens if any of the input tensors have a rank
        higher than 2. Default is None.

    Returns
    -------
    TensorLike or numpy.ndarray or scalar
        An array or a tensor, depending on the inputs.

    Notes
    -----
    For general tensors, the current implementation has an upper limit considering the rank
    of the input tensors. The sum of the ranks of the input tensors plus the sum of contraction
    indices must be at most 26.

    References
    ----------
    https://mathworld.wolfram.com/DotProduct.html

    Examples
    --------
    When working with NumPy arrays, the behaviour coincides with `numpy.dot`. To take the dot
    product of a 2nd order tensor and a vector, use it like this:

    >>> from neumann.linalg import ReferenceFrame, Vector, Tensor2
    >>> from neumann.linalg import dot
    >>> frame = ReferenceFrame(np.eye(3))
    >>> A = Tensor2(np.eye(3), frame=frame)
    >>> v = Vector(np.array([1., 0, 0]), frame=frame)
    >>> dot(A, v)
    Array([1., 0., 0.])

    For general tensors, you have to specify the axes along which contraction happens:

    >>> from neumann.linalg import Tensor
    >>> A = Tensor(np.ones((3, 3, 3, 3)), frame=frame)  # a tensor of order 4
    >>> B = Tensor(np.ones((3, 3, 3)), frame=frame)  # a tensor of order 3
    >>> dot(A, B, axes=(0, 0)).rank
    5
    """
    if isinstance(a, TensorLike) and isinstance(b, TensorLike):
        ra, rb = a.rank, b.rank
        result = None
        if ra == 1 and rb == 1:
            if out is not None:
                raise LinalgOperationInputError(
                    "Parameter 'out' is not allowed with tensors."
                )
            return np.dot(a.show(), b.show())
        elif ra == 2 and rb == 1:
            arr = (a.array @ b.show(a.frame.dual()).T).T
            result = b.__class__(arr, frame=a.frame)
        elif ra == 1 and rb == 2:
            arr = (a.array.T @ b.show(a.frame.dual()).T).T
            result = a.__class__(arr, frame=a.frame)
        elif ra == 2 and rb == 2:
            g = a.frame.Gram()
            result = a.__class__(a.array @ g @ b.show(a.frame), frame=a.frame)
        else:
            if not axes:
                msg = "The parameter 'axes' is required for tensor contraction of general tensors."
                raise LinalgMissingInputError(msg)
            ia = latinrange(ra, start=ord("a"))
            ib = latinrange(rb, start=ord("a") + ra)
            ax_a, ax_b = axes
            ic = latinrange(1, start=ord("a") + ra + rb)[0]
            ia[ax_a] = ic
            ib[ax_b] = ic
            command = "..." + "".join(ia) + "," + "..." + "".join(ib)
            arr = np.einsum(command, a.show(), b.show(), optimize="greedy")
            result = a.__class__._from_any_input(arr)
        if frame:
            result.frame = frame
        return result
    if frame:
        raise LinalgOperationInputError(
            "Parameter 'frame' is exclusive for tensorial inputs."
        )
    if not all([isinstance(x, (ndarray, ArrayWrapper, list)) for x in [a, b]]):
        raise TypeError("Invalid types encountered for dot product.")
    inputs = [x._array if isinstance(x, ArrayWrapper) else x for x in [a, b]]
    return np.dot(*inputs, out=out)


def cross(
    a: Union[TensorLike, ArrayWrapper],
    b: Union[TensorLike, ArrayWrapper],
    *args,
    frame: FrameLike = None,
    **kwargs,
) -> Union[TensorLike, ndarray]:
    """
    Calculates the cross product of two vectors or one vector and a second order
    tensor. The behaviour coincides with NumPy when all inputs are arrays and generalizes
    when they are not, but all inputs must be either all arrays or all tensors of some kind.

    Parameters
    ----------
    *args : Tuple, Optional
        Positional arguments forwarded to NumPy, if all input objects are arrays.
    a : TensorLike or ArrayLike
        A tensor or an array.
    b : TensorLike or ArrayLike
        A tensor or an array.
    frame : FrameLike, Optional
        The target frame of the output. Only if all inputs are TensorLike. If not specified,
        the returned tensor migh be returned in an arbitrary frame, depending on the inputs.
        Default is None.
    **kwargs : dict, Optional
        Keyword arguments forwarded to `numpy.cross`. As of NumPy version '1.22.4', there
        are no keyword arguments for `numpy.cross`, this is to assure compliance with
        all future versions of numpy.

    Returns
    -------
    numpy.ndarray or TensorLike
        An 1d or 2d array, or an 1d or 2d tensor, depending on the inputs.

    References
    ----------
    https://mathworld.wolfram.com/CrossProduct.html

    Examples
    --------
    The cross product of two vectors results in a vector:

    >>> from neumann.linalg import ReferenceFrame, Vector, Tensor2
    >>> from neumann.linalg import cross
    >>> frame = ReferenceFrame(np.eye(3))
    >>> a = Vector(np.array([1., 0, 0]), frame=frame)
    >>> b = Vector(np.array([0, 1., 0]), frame=frame)
    >>> cross(a, b)
    Array([0., 0., 1.])

    The cross product of a second order tensor and a vector result a second order tensor:

    >>> A = Tensor2(np.eye(3), frame=frame)
    >>> cross(A, b)
    Array([[ 0.,  0., -1.],
           [ 0.,  0.,  0.],
           [ 1.,  0.,  0.]])
    """
    if isinstance(a, TensorLike) and isinstance(b, TensorLike):
        ra, rb = a.rank, b.rank
        result = None
        if ra == 1 and rb == 1:
            arr = np.cross(a.array, b.show(a.frame), axis=0)
            result = a.__class__(arr, frame=a.frame)
        elif ra == 2 and rb == 1:
            arr = np.cross(a.show(), b.show(), axis=0)
            result = a.__class__(arr)
        elif ra == 1 and rb == 2:
            arr = np.cross(a.show(), b.show(), axis=0)
            result = b.__class__(arr)
        else:
            msg = (
                "The cross product is not implemented",
                f"for tensors of rank {ra} and {rb}",
            )
            raise NotImplementedError(msg)
        if frame:
            result.frame = frame
        return result
    if frame:
        raise LinalgOperationInputError(
            "Parameter 'frame' is exclusive for tensorial inputs."
        )
    if any([isinstance(x, TensorLike) for x in [a, b]]):
        raise TypeError("Invalid types encountered for dot product.")
    if not all([isinstance(x, (ndarray, ArrayWrapper, list)) for x in [a, b]]):
        raise TypeError("Invalid types encountered for dot product.")
    inputs = [x._array if isinstance(x, ArrayWrapper) else x for x in [a, b]]
    return np.cross(*inputs, *args, **kwargs)


@njit(nogil=True, cache=__cache)
def _show_vector(dcm: np.ndarray, arr: np.ndarray) -> ndarray:
    """
    Returns the coordinates of a single vector in a frame specified
    by a DCM matrix.

    Parameters
    ----------
    dcm : numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    arr : numpy.ndarray
        1d float array of coordinates of a single vector.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the vector with the same shape as `arr`.
    """
    return dcm @ arr


@njit(nogil=True, parallel=True, cache=__cache)
def _show_vectors(dcm: np.ndarray, arr: np.ndarray) -> ndarray:
    """
    Returns the coordinates of multiple vectors in a frame specified
    by a DCM matrix.

    Parameters
    ----------
    dcm : numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    arr : numpy.ndarray
        2d float array of coordinates of multiple vectors.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the vectors with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm @ arr[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _show_vectors_multi(dcm: np.ndarray, arr: np.ndarray) -> ndarray:
    """
    Returns the coordinates of multiple vectors and multiple DCM matrices.

    Parameters
    ----------
    dcm : numpy.ndarray
        The dcm matrix of the transformation as a 3d float array.
    arr : numpy.ndarray
        2d float array of coordinates of multiple vectors.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the vectors with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm[i] @ arr[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _transpose_multi(dcm: np.ndarray) -> ndarray:
    N = dcm.shape[0]
    res = np.zeros_like(dcm)
    for i in prange(N):
        res[i] = dcm[i].T
    return res


def transpose_axes(a: ndarray, axes=None) -> ndarray:
    if len(a.shape) == 2:
        return a.T
    elif len(a.shape) == 3:
        return _transpose_multi(a)
    else:
        return np.transpose(a, axes)


def is_rectangular_frame(axes: ndarray) -> bool:
    """
    Returns True if a frame is Cartesian.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    assert len(axes.shape) == 2, "Input is not a matrix!"
    assert axes.shape[0] == axes.shape[1], "Input is not a square matrix!"
    agram = np.abs(axes @ axes.T)
    return np.isclose(np.trace(agram), np.sum(agram))


def is_normal_frame(axes: ndarray) -> bool:
    """
    Returns True if a frame is normal, meaning, that it's base vectors
    are all of unit length.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return np.allclose(np.linalg.norm(axes, axis=1), 1.0)


def is_orthonormal_frame(axes: ndarray) -> bool:
    """
    Returns True if a frame is orthonormal.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return is_rectangular_frame(axes) and is_normal_frame(axes)


def is_independent_frame(axes: ndarray, tol: float = 0) -> bool:
    """
    Returns True if a the base vectors of a frame are linearly independent.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return np.linalg.det(Gram(axes)) > tol


def is_hermitian(arr: ndarray) -> bool:
    """
    Returns True if the input is a hermitian array.
    """
    shp = arr.shape
    s0 = shp[0]
    return all([s == s0 for s in shp[1:]])


def normalize_frame(axes: ndarray) -> ndarray:
    """
    Returns the frame with normalized base vectors.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return np.array([normalize(a) for a in axes], dtype=axes.dtype)


def Gram(axes: ndarray) -> ndarray:
    """
    Returns the Gram matrix of a frame.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return axes @ axes.T


def dual_frame(axes: ndarray) -> ndarray:
    """
    Returns the dual frame of the input.

    Parameters
    ----------
    axes : numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return transpose_axes(np.linalg.inv(axes))


def is_pos_def(arr) -> bool:
    """
    Returns True if the input is positive definite.
    """
    return np.all(np.linalg.eigvals(arr) > 0)


def is_pos_semidef(arr) -> bool:
    """
    Returns True if the input is positive semi definite.
    """
    return np.all(np.linalg.eigvals(arr) >= 0)


def random_pos_semidef_matrix(N) -> ndarray:
    """
    Returns a random positive semidefinite matrix of shape (N, N).

    Example
    -------
    >>> from neumann.linalg import random_pos_semidef_matrix, is_pos_semidef
    >>> arr = random_pos_semidef_matrix(2)
    >>> is_pos_semidef(arr)
    True
    """
    A = np.random.rand(N, N)
    return A.T @ A


def random_posdef_matrix(N, alpha: float = 1e-12) -> ndarray:
    """
    Returns a random positive definite matrix of shape (N, N).

    All eigenvalues of this matrix are >= alpha.

    Example
    -------
    >>> from neumann.linalg import random_posdef_matrix, is_pos_def
    >>> arr = random_posdef_matrix(2)
    >>> is_pos_def(arr)
    True
    """
    A = np.random.rand(N, N)
    return A @ A.T + alpha * np.eye(N)


def inv_sym_3x3(m: Matrix, as_adj_det=False) -> Matrix:
    P11, P12, P13, P21, P22, P23, P31, P32, P33 = symbols(
        "P_{11} P_{12} P_{13} P_{21} P_{22} P_{23} P_{31} \
                P_{32} P_{33}",
        real=True,
    )
    Pij = [[P11, P12, P13], [P21, P22, P23], [P31, P32, P33]]
    P = sy.Matrix(Pij)
    detP = P.det()
    adjP = P.adjugate()
    invP = adjP / detP
    subs = {s: r for s, r in zip(sy.flatten(P), sy.flatten(m))}
    if as_adj_det:
        return detP.subs(subs), adjP.subs(subs)
    else:
        return invP.subs(subs)


@njit(nogil=True, parallel=True, cache=__cache)
def vpath(p1: ndarray, p2: ndarray, n: int) -> ndarray:
    nD = len(p1)
    dist = p2 - p1
    length = np.linalg.norm(dist)
    s = np.linspace(0, length, n)
    res = np.zeros((n, nD), dtype=p1.dtype)
    d = dist / length
    for i in prange(n):
        res[i] = p1 + s[i] * d
    return res


@njit(nogil=True, cache=__cache)
def linsolve(A, b) -> ndarray:
    return np.linalg.solve(A, b)


@njit(nogil=True, cache=__cache)
def inv(A: ndarray) -> ndarray:
    return np.linalg.inv(A)


@njit(nogil=True, cache=__cache)
def matmul(A: ndarray, B: ndarray) -> ndarray:
    return A @ B


@njit(nogil=True, cache=__cache)
def ATB(A: ndarray, B: ndarray) -> ndarray:
    return A.T @ B


@njit(nogil=True, cache=__cache)
def matmulw(A: ndarray, B: ndarray, w: float = 1.0) -> ndarray:
    return w * (A @ B)


@njit(nogil=True, cache=__cache)
def ATBA(A: ndarray, B: ndarray) -> ndarray:
    return A.T @ B @ A


@njit(nogil=True, cache=__cache)
def ATBAw(A: ndarray, B: ndarray, w: float = 1.0) -> ndarray:
    return w * (A.T @ B @ A)


@guv(["(f8[:, :], f8)"], "(n, n) -> ()", nopython=True, cache=__cache)
def det3x3(A, res):
    res = (
        A[0, 0] * A[1, 1] * A[2, 2]
        - A[0, 0] * A[1, 2] * A[2, 1]
        - A[0, 1] * A[1, 0] * A[2, 2]
        + A[0, 1] * A[1, 2] * A[2, 0]
        + A[0, 2] * A[1, 0] * A[2, 1]
        - A[0, 2] * A[1, 1] * A[2, 0]
    )


@guv(["(f8[:, :], f8)"], "(n, n) -> ()", nopython=True, cache=__cache)
def det2x2(A, res):
    res = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]


@njit(nogil=True, cache=__cache)
def inv2x2(A) -> ndarray:
    res = np.zeros_like(A)
    d = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res[0, 0] = A[1, 1] / d
    res[1, 1] = A[0, 0] / d
    res[0, 1] = -A[0, 1] / d
    res[1, 0] = -A[1, 0] / d
    return res


@guv(["(f8[:, :], f8[:, :])"], "(n, n) -> (n, n)", nopython=True, cache=__cache)
def inv2x2u(A, res):
    d = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res[0, 0] = A[1, 1] / d
    res[1, 1] = A[0, 0] / d
    res[0, 1] = -A[0, 1] / d
    res[1, 0] = -A[1, 0] / d


@guv(["(f8[:, :], f8[:, :])"], "(n, n) -> (n, n)", nopython=True, cache=__cache)
def adj3x3(A, res):
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]


@guv(["(f8[:, :], f8[:, :])"], "(n, n) -> (n, n)", nopython=True, cache=__cache)
def inv3x3u(A, res):
    d = (
        A[0, 0] * A[1, 1] * A[2, 2]
        - A[0, 0] * A[1, 2] * A[2, 1]
        - A[0, 1] * A[1, 0] * A[2, 2]
        + A[0, 1] * A[1, 2] * A[2, 0]
        + A[0, 2] * A[1, 0] * A[2, 1]
        - A[0, 2] * A[1, 1] * A[2, 0]
    )
    res[0, 0] = A[1, 1] * A[2, 2] / d - A[1, 2] * A[2, 1] / d
    res[0, 1] = -A[0, 1] * A[2, 2] / d + A[0, 2] * A[2, 1] / d
    res[0, 2] = A[0, 1] * A[1, 2] / d - A[0, 2] * A[1, 1] / d
    res[1, 0] = -A[1, 0] * A[2, 2] / d + A[1, 2] * A[2, 0] / d
    res[1, 1] = A[0, 0] * A[2, 2] / d - A[0, 2] * A[2, 0] / d
    res[1, 2] = -A[0, 0] * A[1, 2] / d + A[0, 2] * A[1, 0] / d
    res[2, 0] = A[1, 0] * A[2, 1] / d - A[1, 1] * A[2, 0] / d
    res[2, 1] = -A[0, 0] * A[2, 1] / d + A[0, 1] * A[2, 0] / d
    res[2, 2] = A[0, 0] * A[1, 1] / d - A[0, 1] * A[1, 0] / d


@njit(nogil=True, cache=__cache)
def inv3x3(A):
    res = np.zeros_like(A)
    det = (
        A[0, 0] * A[1, 1] * A[2, 2]
        - A[0, 0] * A[1, 2] * A[2, 1]
        - A[0, 1] * A[1, 0] * A[2, 2]
        + A[0, 1] * A[1, 2] * A[2, 0]
        + A[0, 2] * A[1, 0] * A[2, 1]
        - A[0, 2] * A[1, 1] * A[2, 0]
    )
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res /= det
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inv3x3_bulk(A) -> ndarray:
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        det = (
            A[i, 0, 0] * A[i, 1, 1] * A[i, 2, 2]
            - A[i, 0, 0] * A[i, 1, 2] * A[i, 2, 1]
            - A[i, 0, 1] * A[i, 1, 0] * A[i, 2, 2]
            + A[i, 0, 1] * A[i, 1, 2] * A[i, 2, 0]
            + A[i, 0, 2] * A[i, 1, 0] * A[i, 2, 1]
            - A[i, 0, 2] * A[i, 1, 1] * A[i, 2, 0]
        )
        res[i, 0, 0] = A[i, 1, 1] * A[i, 2, 2] - A[i, 1, 2] * A[i, 2, 1]
        res[i, 0, 1] = -A[i, 0, 1] * A[i, 2, 2] + A[i, 0, 2] * A[i, 2, 1]
        res[i, 0, 2] = A[i, 0, 1] * A[i, 1, 2] - A[i, 0, 2] * A[i, 1, 1]
        res[i, 1, 0] = -A[i, 1, 0] * A[i, 2, 2] + A[i, 1, 2] * A[i, 2, 0]
        res[i, 1, 1] = A[i, 0, 0] * A[i, 2, 2] - A[i, 0, 2] * A[i, 2, 0]
        res[i, 1, 2] = -A[i, 0, 0] * A[i, 1, 2] + A[i, 0, 2] * A[i, 1, 0]
        res[i, 2, 0] = A[i, 1, 0] * A[i, 2, 1] - A[i, 1, 1] * A[i, 2, 0]
        res[i, 2, 1] = -A[i, 0, 0] * A[i, 2, 1] + A[i, 0, 1] * A[i, 2, 0]
        res[i, 2, 2] = A[i, 0, 0] * A[i, 1, 1] - A[i, 0, 1] * A[i, 1, 0]
        res[i] /= det
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inv3x3_bulk2(A) -> ndarray:
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        res[i] = inv3x3(A[i])
    return res


@njit(nogil=True, cache=__cache)
def normalize(A) -> ndarray:
    return A / np.linalg.norm(A)


@njit(nogil=True, parallel=True, cache=__cache)
def normalize2d(A) -> ndarray:
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        res[i] = normalize(A[i])
    return res


@njit(nogil=True, cache=__cache)
def norm(A) -> float:
    return np.linalg.norm(A)


@njit(nogil=True, parallel=True, cache=__cache)
def norm2d(A) -> ndarray:
    res = np.zeros(A.shape[0])
    for i in prange(A.shape[0]):
        res[i] = norm(A[i, :])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _linspace(p0: ndarray, p1: ndarray, N):
    s = p1 - p0
    L = np.linalg.norm(s)
    n = s / L
    djac = L / (N - 1)
    step = n * djac
    res = np.zeros((N, p0.shape[0]))
    res[0] = p0
    for i in prange(1, N - 1):
        res[i] = p0 + i * step
    res[-1] = p1
    return res


def linspace(start, stop, N) -> ndarray:
    if isinstance(start, ndarray):
        return _linspace(start, stop, N)
    else:
        return np.linspace(start, stop, N)


@njit(nogil=True, parallel=True, cache=__cache)
def linspace1d(start, stop, N) -> ndarray:
    res = np.zeros(N)
    di = (stop - start) / (N - 1)
    for i in prange(N):
        res[i] = start + i * di
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _transform_tensors2_multi(arr: ndarray, Q: ndarray):
    nE = arr.shape[0]
    res = np.zeros_like(arr)
    for iE in prange(nE):
        res[iE, :, :] = Q[iE] @ arr[iE] @ Q[iE].T
    return res


import numpy as np
from numba import njit, prange


def tr_3333(array: np.ndarray, dcm: np.ndarray):
    Q = dcm.T
    res = np.zeros((3, 3, 3, 3), dtype=array.dtype)
    for p in range(3):
        for q in range(3):
            for r in range(3):
                for s in range(3):
                    for i in range(3):
                        for j in range(3):
                            for k in range(3):
                                for m in range(3):
                                    res[p, q, r, s] += (
                                        Q[p, i]
                                        * Q[q, j]
                                        * Q[r, k]
                                        * Q[s, m]
                                        * array[i, j, k, m]
                                    )
    return res


@njit(nogil=True, parallel=True, cache=True)
def tr_3333_jit(array: np.ndarray, dcm: np.ndarray):
    Q = dcm.T
    res = np.zeros((3, 3, 3, 3), dtype=array.dtype)
    for p in prange(3):
        for q in prange(3):
            for r in prange(3):
                for s in prange(3):
                    for i in prange(3):
                        for j in prange(3):
                            for k in prange(3):
                                for m in prange(3):
                                    res[p, q, r, s] += (
                                        Q[p, i]
                                        * Q[q, j]
                                        * Q[r, k]
                                        * Q[s, m]
                                        * array[i, j, k, m]
                                    )
    return res


def tr_3333_einsum(array: np.ndarray, dcm: np.ndarray, optimize="greedy") -> np.ndarray:
    Q = dcm.T
    return np.einsum("pi, qj, rk, sl, ijkl", Q, Q, Q, Q, array, optimize=optimize)


if __name__ == "__main__":  # pragma: no cover
    from neumann.linalg.frame import ReferenceFrame
    from neumann.linalg.tensor import Tensor, Tensor3333
    import time

    A = ReferenceFrame()
    B = A.orient_new("Body", [0, 0, 90 * np.pi / 180], "XYZ")

    tA = Tensor([[1, 0, 0], [0, 0, 0], [0, 0, 0]], frame=A)
    tA.transform_to_frame(B)

    C = Tensor3333(np.zeros((3, 3, 3, 3)), frame=A)
    dcm = B.dcm()
    array = C._array
    Q = np.transpose(dcm)
    path = np.einsum_path("pi,qj,rk,sl,ijkl", Q, Q, Q, Q, array, optimize="greedy")[0]

    # PERFORMENCE MEASURE
    times = []
    # measuring solution 1
    ts = time.time()
    for i in range(100):
        tr_3333(array, dcm)
    te = time.time()
    times.append((te - ts) * 1000)

    # measuring solution 2
    ts = time.time()
    for i in range(100):
        tr_3333_jit(array, dcm)
    te = time.time()
    times.append((te - ts) * 1000)

    # measuring solution 4
    ts = time.time()
    for i in range(100):
        tr_3333_einsum(array, dcm)
    te = time.time()
    times.append((te - ts) * 1000)

    # measuring solution 5
    ts = time.time()
    for i in range(100):
        tr_3333_einsum(array, dcm, optimize=path)
    te = time.time()
    times.append((te - ts) * 1000)
    print(times)



import numpy as np
from numpy.linalg import norm
from collections.abc import Iterable

from ..function import Function


class MLSWeightFunction(Function):
    def __init__(self, core):
        if not isinstance(core, np.ndarray):
            if isinstance(core, Iterable):
                core = np.array(core)
                dim = core.shape[0]
            elif isinstance(core, int):
                dim = core
                core = np.zeros([dim])
            else:
                raise
        else:
            dim = core.shape[0]
        super().__init__(d=dim)
        self.core = core
        return

    def value(self, x):
        if not isinstance(x, np.ndarray):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                raise
        return


def isMLSWeightFunction(f):
    c1 = isinstance(f, MLSWeightFunction)
    c2 = MLSWeightFunction in list(type(f).__bases__)
    return any([c1, c2])


class ConstantWeightFunction(MLSWeightFunction):
    def __init__(self, dimension, val=1.0):
        super().__init__(np.zeros([dimension]))
        self.val = val
        return

    def value(self, x):
        return self.val

    def gradient(self, x):
        return np.zeros([self.dimension])

    def Hessian(self, x):
        return np.zeros([self.dimension, self.dimension])


class SingularWeightFunction(MLSWeightFunction):
    def __init__(self, core, eps=1e-5):
        super().__init__(core)
        self.eps = eps
        return

    def value(self, x):
        super().value(x)
        return 1 / (norm(np.subtract(self.core, x)) ** 2 + self.eps**2)


class CubicWeightFunction(MLSWeightFunction):
    def __init__(self, core, supportdomain):
        super().__init__(core)
        assert self.dimension == 2
        self.supportdomain = supportdomain
        return

    def inner(self, x):
        d = np.subtract(self.core, x)
        difX = d[0]
        difY = d[1]
        dmX = self.supportdomain[0]
        dmY = self.supportdomain[1]
        rX = abs(difX) / dmX
        rY = abs(difY) / dmY

        if abs(difX) < 1e-12:
            drdX = 0
        else:
            drdX = (difX / abs(difX)) / dmX

        if abs(difY) < 1e-12:
            drdY = 0
        else:
            drdY = (difY / abs(difY)) / dmY

        if rX <= 0.5:
            wX = 2 / 3 - 4 * rX**2 + 4 * rX**3
            dwXdX = (-8 * rX + 12 * rX**2) * drdX
            dwXdXX = (-8 + 24 * rX) * drdX * drdX
        elif rX > 0.5 and rX <= 1:
            wX = 4 / 3 - 4 * rX + 4 * rX**2 - (4 / 3) * rX**3
            dwXdX = (-4 + 8 * rX - 4 * rX**2) * drdX
            dwXdXX = (8 - 8 * rX) * drdX * drdX
        else:
            wX = 0
            dwXdX = 0
            dwXdXX = 0

        if rY <= 0.5:
            wY = 2 / 3 - 4 * rY**2 + 4 * rY**3
            dwYdY = (-8 * rY + 12 * rY**2) * drdY
            dwYdYY = (-8 + 24 * rY) * drdY * drdY
        elif rY > 0.5 and rY <= 1:
            wY = 4 / 3 - 4 * rY + 4 * rY**2 - (4 / 3) * rY**3
            dwYdY = (-4 + 8 * rY - 4 * rY**2) * drdY
            dwYdYY = (8 - 8 * rY) * drdY * drdY
        else:
            wY = 0
            dwYdY = 0
            dwYdYY = 0

        val = wX * wY
        grad = np.array([wY * dwXdX, wX * dwYdY])
        Hessian = np.array([[wY * dwXdXX, dwXdX * dwYdY], [dwXdX * dwYdY, wX * dwYdYY]])

        return val, grad, Hessian

    def value(self, x):
        super().value(x)
        res, _, _ = self.inner(x)
        return res

    def gradient(self, x):
        super().value(x)
        _, res, _ = self.inner(x)
        return res

    def Hessian(self, x):
        super().value(x)
        _, _, res = self.inner(x)
        return res
